/* Copyright (C) 2000-2012 by George Williams */
/*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.

 * The name of the author may not be used to endorse or promote products
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _CONFIG_EXTRA_H
#define _CONFIG_EXTRA_H

// This should have been arranged by the Gnulib configuration.
#if ! defined MALLOC_0_IS_NONNULL || ! MALLOC_0_IS_NONNULL
#error Sorts Mill FontForge 2.0.0_alpha1 requires that malloc(0) return non-NULL.
#endif

// A macro to mark unused function parameters with. We often
// have such parameters, because of extensive use of callbacks.
#if ! defined UNUSED
#if defined __GNUC__
#define UNUSED(x) UNUSED_ ## x _GL_UNUSED
#elif defined __LCLINT__
#define UNUSED(x) /*@unused@*/ x
#else
#define UNUSED(x) x
#endif
#endif // ! UNUSED

/* The iconv encoding to use for the unichar_t type.
   Note that we hope to get rid of that type. */
#ifdef WORDS_BIGENDIAN
# define MY_ICONV_ENCODING_UNICHAR_T MY_ICONV_UCS4BE
# define MY_ICONV_STRING_UNICHAR_T MY_ICONV_UCS4BE_STRING
#else
# define MY_ICONV_ENCODING_UNICHAR_T MY_ICONV_UCS4LE
# define MY_ICONV_STRING_UNICHAR_T MY_ICONV_UCS4LE_STRING
#endif

#if defined HAVE_VISIBILITY && HAVE_VISIBILITY
#define VISIBLE __attribute__((__visibility__("default")))
#else
#define VISIBLE
#endif

#if defined HAVE_TYPEOF
// Override the API-compatible declarations in
// sortsmillff/attributes.h
#define _FF_HAVE_TYPEOF 1
#define _FF_TYPEOF(x) typeof (x)
#define _FF_CAST_TYPEOF(x) (typeof (x))
#endif

#endif // _CONFIG_EXTRA_H
